/* $Id: mimew.c,v 1.5 1998/04/20 16:10:48 moritz Exp $ */
/*
 *  UNIX-Connect, a ZCONNECT(r) Transport and Gateway/Relay.
 *  Copyright (C) 1993-94  Martin Husemann
 *  Copyright (C) 1995     Christopher Creutzig
 *  Copyright (C) 1998     Moritz Both
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 * Eine deutsche Zusammenfassung zum Copyright dieser Programme sowie der
 * GNU General Public License finden Sie in der Datei README.1st.
 * ------------------------------------------------------------------------
 *
 *  Bugreports, suggestions for improvement, patches, ports to other systems
 *  etc. are welcome. Contact the maintainer by e-mail:
 *  christopher@nescio.foebud.org or snail-mail:
 *  Christopher Creutzig, Im Samtfelde 19, 33098 Paderborn
 *
 *  There is a mailing-list for user-support:
 *   unix-connect@mailinglisten.im-netz.de,
 *  write a mail with subject "Help" to
 *   unix-connect-request@mailinglisten.im-netz.de
 *  for instructions on how to join this list.
 */


/*
 * mimew.c - stellt die Routinen fuer MIME zu Verfuegung
 *           (fuer Wandlung ZConnect nach RFC)
 *
 * (C) Copyright 1995 C. Creutzig
 * (C) Copyright 1995-98 M. Both
 */

#include "config.h"
#include <malloc.h>
#include <stdio.h> /* nur fuer sprintf(...) */
#include <string.h>
#include <ctype.h>
#ifdef HAS_STRINGS_H
#include <strings.h>
#endif
#include "lib.h"

/* Zeichen, die in Headerzeilen codiert werden muessen */
static char specialchar[]="()<>@,;:\"/[]?.="; 

/*
 * Diese Funktion ueberprueft, ob ein uebergebener String das MSB gesetzt
 * hast, also Zeichen >127 enthaelt.
 */
int is_8_bit(unsigned char *string)
{
	for(; *string; string++)
	{
		if(*string & 0x80)
		{
			return 1;
		}
	}
	return 0;
}

/*
 * Diese Funktion zaehlt, wieviele Zeichen QP-kodiert werden muessen.
 */
static int count_8_bit(unsigned char *string)
{
	int i;
	
	for(i=0; *string; string++)
	{
		if((*string > 0x7E) || (strchr(specialchar,*string)!=0))
		{
			i++;
		}
	}
	return i;
}

/*
 * diese Funktion konvertiert einen ISO-String nach MIME, falls noetig.
 * Sie setzt ein \t an "Sollbruchstellen" fr Zeilenumbruch ein.
 * start_col ist eine Anzahl von Zeichen, die die erste Zeile krzer
 * wird. Ist sie negativ, findet kein Umbruch statt (den Umbruch brauchen
 * wir lt. RFC).
 */
#define MIME_HD_PFX "=?ISO-8859-1?Q?"
#define MIME_HD_SFX "?="
#define MIME_HD_SIZE (sizeof(MIME_HD_PFX)-1 + sizeof(MIME_HD_SFX)-1)
#define MIME_HD_MAX  (75 - MIME_HD_SIZE)
char *mime_encode(char *iso, int start_col) {
	char *encoded, *enc;
	unsigned char *p;
	int len, col;
				 
	p=(unsigned char*)iso;

	/*
	 * falls noetig, wird konvertiert, ansonsten umkopiert:
	 */
	if(is_8_bit(p))
	{
		/* Laenge des Ausgabestrings: */
		len = strlen(p);
		len += (count_8_bit(p) * 2); /* incl encodedeter Zeichen */
		if (start_col > 0)
			/* Kalkuliere evtl. weiteren Umbruch ein */
			len += start_col;
		/* plus ISO-8859-1 etc., newline, mal Zeilenanzahl;
		 * Zeilenanzahl ergibt sich aus Zeilen-Maximallaenge
		 * von 75 Zeichen excl. newline lt. RFC 2047: */
		len += (MIME_HD_SIZE +2) * (len/(MIME_HD_MAX-2) +1);
		len += 10; /* Sicherheitsabstand */
		encoded = dalloc(len * sizeof(char));
		enc = encoded; /* merken, wo der String anfaengt */

		while(*p) {
			/* Neue codierte Zeile */
			strcpy(encoded, MIME_HD_PFX);
			col = MIME_HD_MAX-2;
			if (start_col > 0) {
				col -= start_col;
				start_col=0;
			}
			encoded += sizeof(MIME_HD_PFX)-1;

			/* Loop, solange die Zeile nicht voll ist (col>0)
			 * oder uns das egal ist (start_col<0)
			 */
			for( ; *p && (col>0 || start_col<0); p++) {
				if (*p == '\t') {
					/* erzwinge Umbruch */
					p++;
					break;
				}
				else 
				if((strchr(specialchar,*p)!=0) || (*p > 0x7E)) {
					sprintf(encoded,"=%02X",(*p));
					encoded+=3;
					col-=3;
				} 
				else if (*p == (char)0x20) {
					*encoded++ = '_';
					col--;
				}
				else {
					*encoded++ = *p;
					col--;
				}
			}
			/* Ende der Zeile */
			*encoded++ = '?';
			*encoded++ = '=';
			/* Hier alle Tabs und andere WS berspringen */
			while (isspace(*p))
				p++;
			if (*p) {
				/* Sollbruchstelle fuer foldputs() */
				*encoded++ = '\t';
			}
		}
		*encoded = '\0';
	}
	else {
		enc = dstrdup(iso);
	}
	
	return enc;
}	
	
/* Funktion zum Konvertieren von 8-Bit-Zeichen in ZConnect-
 * Adressen mit MIME quoted-printable
 */
char *mime_address(char *zcon_ad)
{
	char *mime_ad, *rn, *mime_ad_start;
	char *klammer_auf;
	int len;
	
	len = strlen(zcon_ad) + 2;
	klammer_auf = index(zcon_ad, '(');

	if (!klammer_auf || !is_8_bit(klammer_auf))
	{
		mime_ad = dstrdup(zcon_ad);
		return mime_ad;
	}

	/* Laenge muss groesser werden als 2* zu codierende Zeichen
	 * plus '=?iso-8859-1?q?...?=' - pauschal mit 20 angesetzt.
	 */
	len = len + 2 * count_8_bit(klammer_auf+1) + 20;
	
	mime_ad = dalloc(len * sizeof(char));
	mime_ad_start = mime_ad;
	
	for(; *zcon_ad && !isspace(*zcon_ad) && *zcon_ad !='('; zcon_ad++)
		*mime_ad++=*zcon_ad;

	if(zcon_ad) {
		int klammern, i;
		
		zcon_ad=klammer_auf;
		zcon_ad++;
		*mime_ad++=' ';
		*mime_ad++='(';
		klammer_auf=rindex(zcon_ad,')');
		if(klammer_auf != NULL) *klammer_auf='\0';
		/* Das -1000 verhindert Zeilenumbrueche */
		rn = mime_encode(zcon_ad, -1000);
		strcpy(mime_ad, rn);
		free(rn);
		klammern = count(mime_ad_start,'(') - count(mime_ad_start,')');
		if(klammern<0)
			/* Mehr Klammern zu als auf. ARRGH!
			   Wir ueberschreiben von vorne her
			   ausreichend viele ')' durch ' '.
			   Das ist zwar nicht schoen, erzeugt aber
			   eine gueltige Adresse. */
			for(i=0; i<-klammern; i++)
				*index(mime_ad, ')')=' ';
				/* Wir wissen, da ausreichend viele
				   davon im String sind, brauchen also
				   den Rueckgabewert von index nicht zu pruefen. */
		else
			/* Es fehlen noch Klammern zu. */
			for(i=0; i<klammern && strlen(mime_ad_start) < len; i++)
				strcat(mime_ad,")");
	}else{
		*mime_ad=0;
	}
	return mime_ad_start;
}
